from . import helper



def get_university_name(html_content):
    elements = helper.get_html_elements(html_content, 'h1', 'text-white')
    return elements[0].text


def get_institution_data(html_content, univ_name, about_us_paragraph):
    motto, funding_type = get_university_motto_and_funding_type(html_content)
    res = helper.OrderedDict({
        "name": univ_name,
        "abbreviation": get_university_abbreviation(univ_name),
        "motto": motto,
        "type": funding_type,
        "founding_year": helper.get_answer(
            "When was the University founded?",
            about_us_paragraph
        ),
        "campus_size": helper.get_answer(
            "What is the size of the campus?",
            about_us_paragraph
        ),
        "location": get_location_data(html_content, about_us_paragraph), # ML Inside
        "organization": {
            "schools": helper.get_answer(
                "How many schools does the university have?",
                about_us_paragraph
            ),
            "colleges": helper.get_answer(
                "How many colleges does the university have?",
                about_us_paragraph
            )
        }
    })
    return res


def get_university_abbreviation(univ_name):
    match = helper.re.search(r'\(([^)]+)\)', univ_name)
    if match:
        return match.group(1)
    return 'N/A'


def get_university_motto_and_funding_type(html_content):
    try:
        import sys
        assert int(sys.argv[4]) == 1
    except:
        return 'N/A', 'N/A'

    class_name = 'details'
    tag_name = 'div'
    elements = helper.get_html_elements(html_content, tag_name, class_name)

    paragraph = ''
    for element in elements:
        paragraph += f"\n{element.text}"

    university_motto = 'N/A'
    if not paragraph or not isinstance(paragraph, str) or len(paragraph) == 0:
        return university_motto, 'N/A'

    if 'motto' in paragraph:
        university_motto = helper.get_answer(
            "What is the University's motto?",
            paragraph
        )

    return university_motto, get_university_funding_type(paragraph)

def get_university_funding_type(paragraph):
    if 'public university' in paragraph or\
            'state university' in paragraph or\
            'government university' in paragraph or\
            'publicly funded' in paragraph or\
            'publicly endowed' in paragraph or\
            'state-funded' in paragraph or\
            'government-funded' in paragraph:
        return 'Public'

    if 'private university' in paragraph or\
            'privately funded' in paragraph or\
            'private institution' in paragraph or\
            'private college' in paragraph or\
            'private school' in paragraph or\
            'privately endowed' in paragraph or\
            'independent institution' in paragraph or\
            'independent college' in paragraph or\
            'independent school' in paragraph or\
            'independently funded' in paragraph:
        return 'Private'

    return 'N/A'

def get_location_data(html_content, paragraph):
    class_names = [
        'campus_city',
        'campus_country_code',
        'campus_state'
    ]
    location_data = {
        'city': 'N/A',
        'state': 'N/A',
        'country': 'N/A',
        'features': []
    }
    new_tag_name = 'p'
    for class_name in class_names:
        elements = helper.get_html_elements(html_content, new_tag_name, class_name)
        for element in elements:
            if class_name == 'campus_city':
                location_data['city'] = element.text
            elif class_name == 'campus_country_code':
                location_data['country'] = element.text
            elif class_name == 'campus_state':
                location_data['state'] = element.text

    location_data["features"].append(
        helper.get_answer(
            "Where is the university located, and what is nearby?",
            paragraph
        )
    )

    location_data["features"].append(
        helper.get_answer(
            "What famous locations are mentioned near the university campus?",
            paragraph
        )
    )
    return location_data


from . import helper


def get_available_programs_div(html_content):
    tag_name = 'div'
    class_name = 'avail_prog'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    programs_list_id = 'aptabs'

    try:
        program_list = elements[0].find_all('ul', id=programs_list_id)
        list_elements = program_list[0].find_all('li')
        assert list_elements
    except:
        list_elements = []

    return list_elements


def get_available_programs_data(html_content):
    available_programs = []
    programs_div = get_available_programs_div(html_content)
    if not programs_div:
        return []

    for program in programs_div:
        spans = program.find_all('span')
        try:
            name = spans[0].text
        except:
            name = 'N/A'

        try:
            level = spans[1].text
        except:
            level = 'N/A'

        if name == 'N/A':
            h3s = program.find_all('h3')
            try:
                name = h3s[0].text
            except:
                name = 'N/A'

        if level == 'N/A':
            h3s = program.find_all('h3')
            try:
                level = h3s[1].text
            except:
                level = 'N/A'

        available_programs.append({
            "name": name,
            "level": level,
            "school": 'N/A'
        })

    return available_programs


from multiprocessing import Pool
import logging
import requests
import pickle
import json
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from .utils import *
from . import helper

RETRIES = 3


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

CACHE_FILE = 'ss_scrapping/topuniversities/cachefiles/cache.pkl'
LAST_UPDATED_FILE = 'ss_scrapping/topuniversities/cachefiles/last_updated.txt'

# Load cache from disk
if helper.os.path.exists(CACHE_FILE):
    with open(CACHE_FILE, 'rb') as f:
        cache = pickle.load(f)
else:
    cache = {}

if helper.os.path.exists(LAST_UPDATED_FILE):
    with open(LAST_UPDATED_FILE, 'r') as f:
        last_updated = json.loads(f.read())
else:
    last_updated = {}

chrome_options = Options()
chrome_options.add_argument("--headless")  # Run in headless mode
service = Service('./chromedriver-linux64/chromedriver')  # Update with the path to your ChromeDriver
driver = webdriver.Chrome(service=service, options=chrome_options)


def main():
    if not validate():
        logger.error("Validation failed. Exiting...")
        helper.sys.exit(1)

    total_universities = int(helper.sys.argv[1])
    start_page_number = int(helper.sys.argv[2])
    scrapped_universities = 0
    university_data = []
    while True:
        home_page, url = get_home_page(start_page_number)
        if not home_page:
            logger.error(f"Failed to retrieve the page. Exiting...")
            del cache[url]
            del last_updated[url]
            break

        count_of_universities_on_page = helper.get_count_of_universities_on_page(home_page)
        if not count_of_universities_on_page:
            logger.error(f"Failed to retrieve the count of universities per page. Exiting...")
            del cache[url]
            del last_updated[url]
            break

        university_list = helper.get_university_list(home_page)
        university_count = len(university_list)
        if not university_count:
            logger.error(f"\nFailed to retrieve the list of universities {start_page_number}. Skipping...\n")
            start_page_number += 1
            del cache[url]
            del last_updated[url]
            continue

        scrapped_universities += university_count
        if total_universities < scrapped_universities:
            university_list = university_list[:total_universities]
            university_count = len(university_list)

        if not university_count:
            logger.info("No more universities to scrape")
            break

        data = []
        for university in university_list:
            university_name = university.text
            university_url = university['href']
            if not university_url or not university_name:
                logger.error(f"\nFailed to retrieve the URL or name of the university. Skipping...\n")
                continue
            university_home_page, university_url = get_university_home_page(university_url)
            if not university_home_page:
                logger.error(f"\nFailed to retrieve the page. Skipping...\n")
                continue
            data.append(process_university(university_home_page, university_name, url, university_url))


        university_data.extend(data)
        total_universities -= university_count
        if total_universities <= 0:
            logger.info("Scraped the required number of universities")
            break

        start_page_number = helper.get_next_page_number_from_html_content(home_page)
        if not start_page_number:
            logger.info("No more pages to scrape")
            break

    with open('topuniversities.json', 'w') as f:
        json.dump(university_data, f, indent=4)
    save_cache()
    save_last_updated()
    return


def process_university(university_home_page, university_name, url, university_url):
    about_us_paragraph = get_about_us_paragraph(university_home_page)

    institution_data = get_institution_data(university_home_page, university_name, about_us_paragraph)
    student_data = get_student_data(university_home_page)
    faculty_data = get_faculty_data(university_home_page)
    admission_requirements = get_admission_data(university_home_page)
    available_programs = get_available_programs_data(university_home_page)
    ratings = get_ranking_data(university_home_page)
    financial_information = get_financial_data(university_home_page)
    meta_data = get_metadata(university_home_page, last_updated.get(university_url, 'N/A'))

    # with Pool() as pool:
    #     tasks = [
    #         pool.apply_async(get_institution_data, (university_home_page, university_name, about_us_paragraph)),
    #         pool.apply_async(get_student_data, (university_home_page,)),
    #         pool.apply_async(get_faculty_data, (university_home_page,)),
    #         pool.apply_async(get_admission_data, (university_home_page,)),
    #         pool.apply_async(get_available_programs_data, (university_home_page,)),
    #         pool.apply_async(get_ranking_data, (university_home_page,)),
    #         pool.apply_async(get_financial_data, (university_home_page,)),
    #         pool.apply_async(get_metadata, (university_home_page, last_updated.get(university_url, 'N/A'))),
    #     ]

    #     results = [task.get() for task in tasks]

    # institution_data, student_data, faculty_data, admission_requirements, available_programs, \
    # ratings, financial_information, meta_data = results

    parallel_results = {
        "schools_answer": get_answer_in_list("What are the schools and the college at the university?", about_us_paragraph),
        "varsity_teams_answer": get_answer("How many varsity sports does the institution boast?", about_us_paragraph),
        "student_participation_answer": get_answer("What percentage of undergraduates join a sports team?", about_us_paragraph),
        "museums_galleries_answer": get_answer("How many museums and galleries are on the campus?", about_us_paragraph),
        "museum_visitors_answer": get_answer("How many visitors does the Museum draw?", about_us_paragraph),
        "total_answer": get_answer("How many music, theatre, writing, and dance groups do students participate in?", about_us_paragraph),
        "categories_answer": get_answer("What art forms do students participate in at the university?", about_us_paragraph),
        "current_areas_answer": get_answer("What areas are researchers at the forefront of developments in?", about_us_paragraph),
        "radar_breakthrough": get_answer("What scientific breakthrough did research from this institution contribute to related to radar?", about_us_paragraph),
        "core_memory_breakthrough": get_answer("What scientific breakthrough did research from the institution contribute to related to magnetic core memory?", about_us_paragraph),
        "universe_breakthrough": get_answer("What scientific breakthrough did research from the institution contribute to related to the expanding universe?", about_us_paragraph),
        "total_active_answer": get_answer("How many active companies have alumni launched?", about_us_paragraph),
        "jobs_created_answer": get_answer("How many jobs were created by alumni?", about_us_paragraph),
        "annual_revenue_answer": get_answer("How much annual revenue is generated by alumni's companies?", about_us_paragraph),
    }

    # tasks_to_parallelize = {
    #     "schools_answer": ("get_answer_in_list", "What are the schools and the college at the university?"),
    #     "varsity_teams_answer": ("get_answer", "How many varsity sports does the institution boast?"),
    #     "student_participation_answer": ("get_answer", "What percentage of undergraduates join a sports team?"),
    #     "museums_galleries_answer": ("get_answer", "How many museums and galleries are on the campus?"),
    #     "museum_visitors_answer": ("get_answer", "How many visitors does the Museum draw?"),
    #     "total_answer": ("get_answer", "How many music, theatre, writing, and dance groups do students participate in?"),
    #     "categories_answer": ("get_answer", "What art forms do students participate in at the university?"),
    #     "current_areas_answer": ("get_answer", "What areas are researchers at the forefront of developments in?"),
    #     "radar_breakthrough": ("get_answer", "What scientific breakthrough did research from this institution contribute to related to radar?"),
    #     "core_memory_breakthrough": ("get_answer", "What scientific breakthrough did research from the institution contribute to related to magnetic core memory?"),
    #     "universe_breakthrough": ("get_answer", "What scientific breakthrough did research from the institution contribute to related to the expanding universe?"),
    #     "total_active_answer": ("get_answer", "How many active companies have alumni launched?"),
    #     "jobs_created_answer": ("get_answer", "How many jobs were created by alumni?"),
    #     "annual_revenue_answer": ("get_answer", "How much annual revenue is generated by alumni's companies?"),
    # }

    # with Pool() as pool:
    #     async_results = {
    #         key: pool.apply_async(globals()[func], (query, about_us_paragraph))
    #         for key, (func, query) in tasks_to_parallelize.items()
    #     }

    #     parallel_results = {key: result.get() for key, result in async_results.items()}

    historical_achievements_answer = ([
        parallel_results["radar_breakthrough"],
        parallel_results["core_memory_breakthrough"],
        parallel_results["universe_breakthrough"]
    ])

    return {
        "institution": institution_data, # ML Inside
        "student_body": student_data,
        "faculty": faculty_data,
        "admission_requirements": admission_requirements,
        "programs": {
            "schools": parallel_results["schools_answer"],
            "available_programs": available_programs
        },
        "campus_life": {
            "sports": {
                "varsity_teams": parallel_results["varsity_teams_answer"],
                "student_participation": parallel_results["student_participation_answer"]
            },
            "arts_culture": {
                "museums_galleries": parallel_results["museums_galleries_answer"],
                "museum_visitors": parallel_results["museum_visitors_answer"],
                "student_groups": {
                    "total": parallel_results["total_answer"],
                    "categories": [parallel_results["categories_answer"]]
                }
            }
        },
        "rankings": ratings,
        "research_focus": {
            "current_areas": [parallel_results["current_areas_answer"]],
            "historical_achievements": historical_achievements_answer
        },
        "alumni_impact": {
            "companies": {
                "total_active": parallel_results["total_active_answer"],
                "jobs_created": parallel_results["jobs_created_answer"],
                "annual_revenue": parallel_results["annual_revenue_answer"]
            }
        },
        "financial_information": financial_information,
        "metadata": meta_data,
        "urls": {
            "home_page_url": url,
            "university_url": university_url
        }
    }

def validate():
    """
    Validate the command line arguments
    """
    if len(helper.sys.argv) not in [4, 5]:
        logger.error("Please provide the number of universities to scrape and the start page number")
        return False

    try:
        count_of_universities = int(helper.sys.argv[1])
    except Exception as e:
        logger.error(f"Please provide a valid number of universities {e}")
        return False

    try:
        start_page_number = int(helper.sys.argv[2])
        assert start_page_number >= 1
    except Exception as e:
        logger.error(f"Please provide a valid choice for start page number {e}")
        return False

    try:
        assert count_of_universities >= 1 and count_of_universities <= 1500
    except AssertionError as e:
        logger.error(f"Please provide the number of universities between 1 and 500 {e}")
        return False

    try:
        use_cache = int(helper.sys.argv[3])
        assert use_cache == 0 or use_cache == 1
    except Exception as e:
        logger.error(f"Please provide a valid choice for cache {e}")
        return False

    return True


def get_home_page(page, retries=0):
    url = f"https://www.topuniversities.com/world-university-rankings?page={page}"
    home_page = get_html_content(url)
    if not home_page and retries < RETRIES:
        logger.error(f"Failed to retrieve the page: {url}. Retrying...")
        return get_home_page(page, retries + 1)
    return home_page, url


def get_university_home_page(university_href, retries=0):
    university_home_page = None
    try:
        url = 'https://www.topuniversities.com' + university_href
        university_home_page = get_html_content(url, university_home=True)
    except Exception as e:
        if retries < RETRIES:
            logger.error(f"Failed to retrieve the page: {url}. Retrying...")
            return get_university_home_page(university_href, retries + 1)

    if not university_home_page and retries < RETRIES:
        logger.error(f"Failed to retrieve the page: {url}. Retrying...")
        return get_university_home_page(university_href, retries + 1)
    return university_home_page, url


def get_html_content(url, university_home=False):
    if int(helper.sys.argv[3]) and url in cache:
        logger.info(f"Using cached content for URL: {url}")
        return cache[url]

    response = requests.head(url)
    if response.status_code != 200:
        logger.error(f"Failed to retrieve the page. Status code: {response.status_code}")
        return None

    last_modified = response.headers.get('Last-Modified', None)
    if last_modified and last_updated.get(url) == last_modified:
        logger.info(f"Using cached content for URL: {url} as it has not been modified")
        return cache[url]

    logger.info(f"Fetching live content for URL: {url}")
    try:
        driver.get(url)
        if university_home:
            WebDriverWait(driver, 20).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, '.circle'))
            )
        else:
            WebDriverWait(driver, 20).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, '.uni-link'))
            )
        html_content = driver.page_source
    except Exception as e:
        try:
            driver.get(url)
            driver.implicitly_wait(20)
            html_content = driver.page_source
        except Exception as e:
            logger.error(f"Failed to asretrieve the page: {url}. Error: {e}")
            return None

    cache[url] = html_content
    last_updated[url] = last_modified
    return html_content


def save_cache():
    with open(CACHE_FILE, 'wb') as f:
        pickle.dump(cache, f)
    return

def save_last_updated():
    with open(LAST_UPDATED_FILE, 'w') as f:
        f.write(json.dumps(last_updated))
    return


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        logger.error(f"An error occurred {e}. Exiting...")
    finally:
        driver.quit()
        helper.sys.exit(0)






from . import helper

def has_info(html_content):
    tag_name = 'h2'
    id_attr = 'cost'
    element = helper.get_html_element_by_id(html_content, tag_name, id_attr)
    if not element:
        return False
    return True


def get_financial_data(html_content):
    data = {
        "tuition_and_fees": 'N/A',
        "room_and_board": 'N/A',
        "books_and_supplies": 'N/A',
        "financial_aid": {
            "students_receiving_aid": 'N/A',
            "avg_need_based_aid": 'N/A',
            "avg_need_met": 'N/A',
        }
    }

    fin_div = get_financial_details_div(html_content)
    inner_divs = fin_div.find_all('div', class_='ContentSection__StyledDataRow-sc-699pa9-4')
    for div in inner_divs:
        paragraphs = div.find_all('p', class_='Paragraph-sc-1iyax29-0')
        if not paragraphs or len(paragraphs) < 2:
            continue

        if 'Tuition' in paragraphs[0].text:
            data['tuition_and_fees'] = paragraphs[1].text
        elif 'Food & Housing' in paragraphs[0].text:
            data['room_and_board'] = paragraphs[1].text
        elif 'Average Need-Based Aid Package' in paragraphs[0].text:
            data['financial_aid']['avg_need_based_aid'] = paragraphs[1].text

    return data


def get_financial_details_div(html_content):
    h2_element = helper.get_html_element_by_id(html_content, 'h2', 'cost')
    if h2_element:
        next_div = h2_element.find_next_sibling('div', class_='Grid-lx2f3i-0 ContentSection__StyledGrid-sc-699pa9-1 fQYHzS eWFIVC sm-hide')
        if next_div:
            return next_div.find('div', class_='ContentSection__StyledCell-sc-699pa9-3')

    return None



from . import helper


def get_faculty_data(html_content):
    percents = get_staff_percents(html_content)
    res = helper.OrderedDict({
        "total_staff": get_staff_count(html_content),
        "composition": {
            "domestic": {
                "percentage": percents['domestic_percent']
            },
            "international": {
                "percentage": percents['international_percent']
            }
        }
    })
    return res


def get_staff_count(html_content):
    class_name = 'studstaff-subsection-count'
    tag_name = 'div'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    try:
        staff_count = elements[2].text
    except:
        staff_count = 'N/A'
    return staff_count


def get_staff_percents(html_content):
    parent_tag = 'div'
    parent_class = 'color-code-parent'
    nested_tag = 'div'

    elements = helper.get_html_elements(html_content, parent_tag, parent_class)

    try:
        domestic_percent = helper.get_innermost_nested_element(elements[4], nested_tag)
    except:
        domestic_percent = 'N/A'

    try:
        international_percent = helper.get_innermost_nested_element(elements[5], nested_tag)
    except:
        international_percent = 'N/A'

    res = {
        "domestic_percent": domestic_percent,
        "international_percent": international_percent,
    }

    return res


from . import helper


def data_exists(html_content):
    try:
        h2_element = helper.get_html_element_by_id(html_content, 'h2', 'aftercollege')
        assert h2_element
    except:
        return False
    return True


def get_median_earnings_6_years(html_content):
    median_earnings_6_years = 'N/A'
    try:
        h2_element = helper.get_html_element_by_id(html_content, 'h2', 'aftercollege')
        if h2_element:
            next_div = h2_element.find_next_sibling('div', class_='Grid-lx2f3i-0 ContentSection__StyledGrid-sc-699pa9-1 fQYHzS eWFIVC sm-hide')
            if next_div:
                median_earnings_6_years = next_div.find('p', class_='AfterCollegeSection__DataHeader-czoza-1 iYbKOL').text.lstrip('$')
    except:
        pass
    return median_earnings_6_years


def get_student_debt_div(html_content):
    try:
        h2_element = helper.get_html_element_by_id(html_content, 'h2', 'aftercollege')
        if h2_element:
            next_div = h2_element.find_next_sibling('div', class_='Grid-lx2f3i-0 ContentSection__StyledGrid-sc-699pa9-1 fQYHzS eWFIVC sm-hide')
            if next_div:
                return next_div.find('div')
    except:
        return 'N/A'

def get_after_graduation_data(html_content):
    data = {
        "salary_stats": {
            "median_earnings_6_years": 'N/A',
            "median_earnings_10_years": "N/A",
        },
        "employment_stats": {
            "employment_rate_2_years": "N/A",
        },
        "grad_school_stats": {
            "grad_school_attendance_rate": "N/A",
        },
        "student_debt": {
            "graduates_with_debt": "N/A",
            "avg_debt_at_graduation": "N/A",
            "loan_default_rate": "N/A",
        }
    }

    if not data_exists(html_content):
        return data

    data["salary_stats"]["median_earnings_6_years"] = get_median_earnings_6_years(html_content)

    student_debt_div = get_student_debt_div(html_content)
    student_debt_divs = student_debt_div.find_all('div')
    # print(student_debt_divs)

    try:
        paragraphs = student_debt_divs[1].find_all('p')
        if paragraphs and len(paragraphs) >= 2:
            data["student_debt"]["graduates_with_debt"] = paragraphs[1].text.rstrip('%')
    except:
        pass

    try:
        paragraphs = student_debt_divs[2].find_all('p')
        if paragraphs and len(paragraphs) >= 2:
            data["student_debt"]["avg_debt_at_graduation"] = paragraphs[1].text.lstrip('$')
    except:
        pass
    return data


from . import helper


# def get_available_programs_div(html_content):
#     tag_name = 'div'
#     class_name = 'avail_prog'
#     elements = helper.get_html_elements(html_content, tag_name, class_name)
#     programs_list_id = 'aptabs'

#     try:
#         program_list = elements[0].find_all('ul', id=programs_list_id)
#         list_elements = program_list[0].find_all('li')
#         assert list_elements
#     except:
#         list_elements = []

#     return list_elements


def get_available_programs_data(html_content):
    data = {
        "student_faculty_ratio": "N/A",
        "class_stats": {
            "under_20_students": "N/A",
            "under_50_students": "N/A",
        },
        "most_popular_majors": [

        ],
        "graduation_rates": {
            "4_year": "N/A",
            "6_year": "N/A",
        },
        "retention_rate": "N/A",
    }

    div_elements = helper.get_html_elements(html_content, 'p', 'AcademicSection__DataHeader-sc-1g5x16k-5')
    if div_elements:
        try:
            data["graduation_rates"]["4_year"] = div_elements[0].text
        except:
            pass

        try:
            data["student_faculty_ratio"] = div_elements[1].text
        except:
            pass

        try:
            data["class_stats"]["under_20_students"] = div_elements[2].text
        except:
            pass

    popular_major_div = helper.get_html_elements(html_content, 'div', 'AcademicSection__StyledDataRow-sc-1g5x16k-12')
    for div in popular_major_div:
        paragraphs = div.find_all('p')
        if paragraphs and len(paragraphs) >= 2:
            data["most_popular_majors"].append(f"{paragraphs[0].text} ({paragraphs[1].text})")

    return data




from . import helper
from . import institution
from . import student
from . import faculty
from . import admission_req
from . import programs
from . import rankings
from . import fin_info
from . import metadata



def get_about_us_paragraph(university_home_page):
    return helper.get_about_us_paragraph(university_home_page)

def get_institution_data(university_home_page, university_name, about_us_paragraph):
    return institution.get_institution_data(university_home_page, university_name, about_us_paragraph)

def get_student_data(university_home_page):
    return student.get_student_data(university_home_page)

def get_faculty_data(university_home_page):
    return faculty.get_faculty_data(university_home_page)

def get_admission_data(university_home_page):
    return admission_req.get_admission_data(university_home_page)

def get_available_programs_data(university_home_page):
    return programs.get_available_programs_data(university_home_page)

def get_ranking_data(university_home_page):
    return rankings.get_ranking_data(university_home_page)

def get_financial_data(university_home_page):
    return fin_info.get_financial_data(university_home_page)

def get_metadata(university_home_page, last_updated):
    return metadata.get_metadata(university_home_page, last_updated)

def get_answer_in_list(question, about_us_paragraph):
    return helper.get_answer_in_list(question, about_us_paragraph)

def get_answer(question, about_us_paragraph):
    return helper.get_answer(question, about_us_paragraph)

# def get_answer_2(question, about_us_paragraph):
#     return helper.get_answer_2(question, about_us_paragraph)


from bs4 import BeautifulSoup
from collections import OrderedDict
import logging
import sys
import re

try:
    assert int(sys.argv[4]) == 1
except:
    pass
else:
    from transformers import pipeline

import os


NLP_MODEL = os.getenv("NLP_MODEL")
# NLP_MODEL = "distilbert/distilbert-base-cased-distilled-squad"
# NLP_MODEL = "bert-large-uncased-whole-word-masking-finetuned-squad"

logger = logging.getLogger(__name__)


def get_html_elements(html_content, tag_name, class_name):
    elements = None
    try:
        soup = BeautifulSoup(html_content, 'html.parser')
        elements = soup.find_all(tag_name, class_=class_name)
    except Exception as e:
        logger.error(f"Error while parsing the html content: {e}")
    return elements

def get_innermost_nested_element(parent, nested_tag):
    nested_elements = parent.find_all(nested_tag)
    return (nested_elements[-1].text).rstrip('%')

def get_html_element_by_id(html_content, tag_name, id_name):
    soup = BeautifulSoup(html_content, 'html.parser')
    element = soup.find(tag_name, id=id_name)
    return element

def get_next_page(unordered_list, list_element):
    list_elements = unordered_list.find_all(list_element)
    list_element_anchor_tag = list_elements[-1].find('a')
    if not list_element_anchor_tag:
        return None
    next_page = list_element_anchor_tag['href'].lstrip('#')
    return next_page

def get_next_page_number_from_html_content(html_content):
    unordered_list = get_html_element_by_id(html_content, 'ul', 'alt-style-pagination')
    next_page = get_next_page(unordered_list, 'li')
    # next_page = page-2
    try:
        assert next_page
        next_page_number = int(next_page.split('-')[-1])
    except:
        return None
    return next_page_number

def get_count_of_universities_on_page(html_content):
    try:
        results_per_page = get_html_element_by_id(html_content, 'input', 'perpagedata').get('value')
        results_per_page = int(results_per_page)
    except:
        return 0
    return results_per_page

def get_university_list(html_content):
    tag_name = 'a'
    class_name = 'uni-link'
    elements = get_html_elements(html_content, tag_name, class_name=class_name)
    return elements

def get_about_us_paragraph(html_content):
    class_name = 'details'
    tag_name = 'div'
    elements = get_html_elements(html_content, tag_name, class_name)

    paragraph = ''
    for element in elements:
        paragraph += f"\n{element.text}"

    return paragraph

def get_answer(question, paragraph):
    # print(question)
    # if question in [
    #     'How many active companies have alumni launched?',
    #     'How many jobs were created by alumni?',
    #     "How much annual revenue is generated by alumni's companies?",
    #     "What scientific breakthrough did research from the institution contribute to related to the expanding universe?"
    # ]:
    #     return 'N/A'
    try:
        import sys
        assert int(sys.argv[4]) == 1
        assert paragraph and isinstance(paragraph, str) and len(paragraph) > 0
    except:
        return 'N/A'

    qa_pipeline = pipeline(
        "question-answering",
        model=NLP_MODEL
    )
    return qa_pipeline(
        question=question,
        context=paragraph
    ).get('answer', None)


def get_answer_in_list(question, paragraph):
    try:
        import sys
        assert int(sys.argv[4]) == 1
    except:
        return ['N/A']

    try:
        answers = get_answer(question, paragraph)
        assert answers
        answers = [answers]
    except:
        answers = []

    return answers


# def check_page_end(html_content):
#     tag_name = 'div'
#     class_name = 'no-ranking-results-found'

#     elements = helper.get_html_elements(html_content, tag_name, class_name)
#     return bool(elements)



from . import helper

def is_scholarships_available(html_content):
    div = helper.get_html_element_by_id(html_content, 'div', 'p2-tuition-fee-and-scholarships')
    if not div:
        return False
    return True


def get_features(html_content):
    tag_name = 'p'
    class_name = 'guide-keypoints'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    features = []
    for element in elements:
        try:
            feature = element.text
        except:
            feature = 'N/A'
        features.append(feature)
    return features

def get_financial_data(html_content):
    data = {
        "scholarships": {
            "available": is_scholarships_available(html_content),
            "guide": {
                "features": get_features(html_content)
            }
        }
    }

    return data




from . import helper


def get_metadata(html_content, last_updated):
    content = helper.get_html_element_by_id(html_content, 'p', 'wur-content')

    year = 'N/A'
    data_source = 'QS World University Rankings'
    if content and data_source in content.text:
        year = extract_year(content.text) or 'N/A'
    else:
        data_source = 'N/A'

    return {
        "data_source": data_source,
        "last_updated": last_updated,
        "ranking_year": year
    }

def extract_year(text):
    match = helper.re.search(r'QS World University Rankings (\d{4})', text)
    if match:
        return match.group(1)
    return None


from . import helper


def get_notable_features(html_content, about_us_paragraph):
    data = {
        "specialized_programs": [
            helper.get_answer(
                "What are the names of the highly ranked graduate programs at the university?",
                about_us_paragraph

            )
        ],
        "research": get_faculty_resesrch(html_content),
        "notable_alumni": [
            helper.get_answer(
                "Who are some of the notable alumni of the university?",
                about_us_paragraph
            )
        ]
    }

    return data
#

def get_faculty_resesrch(html_content):
    research = {
        "research_expenditures": 'N/A',
        "bibliometric_rank": 'N/A',
        "total_papers": 'N/A',
        "citations_per_publication": 'N/A',
        "field_weighted_citation_impact": 'N/A',
    }

    research_div = helper.get_html_elements(html_content, 'div', 'FacultyResearchImpact__StyledDataRow-ewzxnx-2')
    if research_div and len(research_div) > 0:
        for div in research_div:
            paragraphs = div.find_all('p')
            if paragraphs and len(paragraphs) >= 2:
                key = paragraphs[0].text
                value = paragraphs[1].text

                if key == "Bibliometric Rank":
                    research["bibliometric_rank"] = value
                if "Total Papers published" in key:
                    research["total_papers"] = value
                if key == "Citations Per Publication":
                    research["citations_per_publication"] = value
                if key == "Field Weighted Citation Impact":
                    research["field_weighted_citation_impact"] = value

    return research


from multiprocessing import Pool
import logging
import requests
import pickle
import json
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from .utils import *
from . import helper

RETRIES = 3


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

CACHE_FILE = 'ss_scrapping/usnews/cachefiles/cache.pkl'
LAST_UPDATED_FILE = 'ss_scrapping/usnews/cachefiles/last_updated.txt'

# Load cache from disk
if helper.os.path.exists(CACHE_FILE):
    with open(CACHE_FILE, 'rb') as f:
        cache = pickle.load(f)
else:
    cache = {}

if helper.os.path.exists(LAST_UPDATED_FILE):
    with open(LAST_UPDATED_FILE, 'r') as f:
        last_updated = json.loads(f.read())
else:
    last_updated = {}

chrome_options = Options()
chrome_options.add_argument("User-Agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36")
chrome_options.add_argument("--headless=new")
chrome_options.add_argument("--disable-blink-features=AutomationControlled")
chrome_options.add_argument("--disable-infobars")
chrome_options.add_argument("--start-maximized")
chrome_options.add_argument("--disable-gpu")
chrome_options.add_argument("--disable-dev-shm-usage")
chrome_options.add_argument("--no-sandbox")
chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])
chrome_options.add_experimental_option("useAutomationExtension", False)
service = Service('./chromedriver-linux64/chromedriver')
driver = webdriver.Chrome(service=service, options=chrome_options)


def main():
    if not validate():
        logger.error("Validation failed. Exiting...")
        helper.sys.exit(1)

    home_page, url = get_home_page()
    if not home_page:
        logger.error(f"Failed to retrieve the page. Exiting...")
        del cache[url]
        del last_updated[url]
        save_cache()
        save_last_updated()
        return

    count_of_universities_on_homepage = helper.get_count_of_universities_on_page(home_page)
    count_of_university_elements_in_list = helper.get_universities_count(home_page)

    if count_of_universities_on_homepage != count_of_university_elements_in_list:
        logger.info(f"Count of universities on homepage: {count_of_universities_on_homepage} "
                     f"does not match with the count of universities in the list: {count_of_university_elements_in_list}.")
        logger.info("Retrieving the list of universities from the homepage.")

    # Anchor tags containing the university's url and nested name
    university_list = helper.get_university_list(home_page)
    university_data = []
    c = len(university_list)
    for count, university in enumerate(university_list):
        university_name, university_url = helper.get_university_name_and_url(university)
        if not university_url or not university_name:
            logger.error(f"\nFailed to retrieve the URL or name of the university. Skipping...\n")
            continue

        university_home_page, university_url = get_university_home_page(university_url)
        if not university_home_page:
            logger.error(f"\nFailed to retrieve the page. Skipping...\n")
            continue
        else:
            logger.info(f"\nSuccessfully retrieved the page: {university_url}")
            logger.info(f"{count + 1} of {c} universities processed.\n")

        university_data.append(process_university(university_home_page, university_name, url, university_url))



    with open('usnews.json', 'w') as f:
        json.dump(university_data, f, indent=4)
    save_cache()
    save_last_updated()
    return


def process_university(university_home_page, university_name, url, university_url):
    about_us_paragraph = get_about_us_paragraph(university_home_page)

    institution_data = get_institution_data(university_home_page, university_name, about_us_paragraph)
    student_data = get_student_data(university_home_page, about_us_paragraph)
    admission_data = get_admission_data(university_home_page)
    academics_data = get_available_programs_data(university_home_page)
    rankings_data = get_ranking_data(university_home_page)
    financial_data = get_financial_data(university_home_page)
    after_graduation_data = get_after_graduation_data(university_home_page)
    features_data = get_notable_features(university_home_page, about_us_paragraph)
    meta_data = get_metadata(university_home_page, last_updated.get(university_url, 'N/A'))

    # with Pool() as pool:
    #     tasks = [
    #         pool.apply_async(get_institution_data, (university_home_page, university_name, about_us_paragraph)),
    #         pool.apply_async(get_student_data, (university_home_page, about_us_paragraph)),
    #         pool.apply_async(get_admission_data, (university_home_page,)),
    #         pool.apply_async(get_available_programs_data, (university_home_page,)),
    #         pool.apply_async(get_ranking_data, (university_home_page,)),
    #         pool.apply_async(get_financial_data, (university_home_page,)),
    #         pool.apply_async(get_after_graduation_data, (university_home_page,)),
    #         pool.apply_async(get_notable_features, (university_home_page, about_us_paragraph)),
    #         pool.apply_async(get_metadata, (university_home_page, last_updated.get(university_url, 'N/A'))),
    #     ]

    #     results = [task.get() for task in tasks]

    # institution_data, student_data, admission_data, academics_data, rankings_data, financial_data,  \
    #     after_graduation_data, features_data, meta_data = results

    return {
        "basic_info": institution_data,
        "rankings": rankings_data,
        "admissions": admission_data,
        "costs_and_aid": financial_data,
        "academics": academics_data,
        "student_life": student_data,
        "after_graduation": after_graduation_data,
        "notable_features": features_data,
        "metadata": meta_data,
        "url": university_url
    }

def validate():
    """
    Validate the command line arguments
    """
    if len(helper.sys.argv) not in [2, 3]:
        logger.error("Please provide the two command line arguments. Exiting...")
        return False

    try:
        use_cache = int(helper.sys.argv[1])
    except Exception as e:
        logger.error(f"Please provide a valid number of Cache choice: {e}")
        return False

    return True


def get_home_page(retries=0):
    url = f"https://www.usnews.com/best-colleges/rankings/national-universities"
    home_page = get_html_content(url)
    if not home_page and retries < RETRIES:
        logger.error(f"Failed to retrieve the page: {url}. Retrying...")
        return get_home_page(retries + 1)
    logger.info(f"Successfully retrieved the home page: {url}")
    return home_page, url


def get_university_home_page(university_href, retries=0):
    university_home_page = None
    try:
        url = 'https://www.usnews.com' + university_href
        university_home_page = get_html_content(url, university_home=True)
    except Exception as e:
        if retries < RETRIES:
            logger.error(f"Failed to retrieve the page: {url}. Retrying...")
            return get_university_home_page(university_href, retries + 1)

    if not university_home_page and retries < RETRIES:
        logger.error(f"Failed to retrieve the page: {url}. Retrying...")
        return get_university_home_page(university_href, retries + 1)
    return university_home_page, url

def scroll_to_load_content(driver, scroll_top=0):
    scroll_pause_time = 1
    while True:
        scroll_height = driver.execute_script("return document.body.scrollHeight")
        scroll_increment = int((scroll_height - scroll_top) / 4) or 10
        logger.info("Scrolling to the bottom of the page")

        for i in range(scroll_top, scroll_height, scroll_increment):
            driver.execute_script(f"window.scrollTo({i}, {i + scroll_increment});")
            helper.time.sleep(scroll_pause_time)

        try:
            # anchor tags containing the university's url and nested name
            university_elements = driver.find_elements(By.CSS_SELECTOR, '.Card__StyledAnchor-sc-1ra20i5-10')
        except Exception as e:
            logger.error(f"Failed to retrieve the list of universities: {e}")
        else:
            logger.info(f"Found {len(university_elements)} universities.")

        offsetHeight = driver.execute_script("return document.body.offsetHeight")
        if offsetHeight == scroll_height:
            break
        scroll_top = scroll_height


    try:
        button = WebDriverWait(driver, 20).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'pager__ButtonStyled-sc-1i8e93j-1'))
        )
        driver.execute_script("arguments[0].scrollIntoView(true);", button)  # Scroll the button into view
        WebDriverWait(driver, 20).until(EC.element_to_be_clickable((By.CLASS_NAME, 'pager__ButtonStyled-sc-1i8e93j-1')))
        button.click()
        helper.time.sleep(1)
    except Exception as e:
        logger.error(f"Failed to click the button: {e}. Stopping the scroll.")
        return driver.page_source
    return scroll_to_load_content(driver, scroll_top=scroll_top)


def get_html_content(url, university_home=False):
    if int(helper.sys.argv[1]) and url in cache:
        logger.info(f"Using cached content for URL: {url}")
        return cache[url]

    last_modified = None

    headers = {
        "User-Agent":
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    }
    try:
        response = requests.get(url, timeout=10, headers=headers)
        if response.status_code != 200:
            logger.error(f"Failed to retrieve the page. Status code: {response.status_code}")
            return None
    except requests.exceptions.Timeout:
        logger.error("Request timed out")
        return None
    except requests.exceptions.RequestException as e:
        logger.error(f"Request failed: {e}")
        return None

    last_modified = response.headers.get('Date', None)

    if last_modified and last_updated.get(url) == last_modified:
        logger.info(f"No update found. Using cached content for URL: {url}")
        return cache[url]

    logger.info(f"Fetching live content for URL: {url}")
    if university_home:

        try:
            driver.get(url)
            WebDriverWait(driver, 5).until(
                EC.presence_of_element_located((By.CSS_SELECTOR, '.pr-snippet-review-count, .ContentSection__Header-sc-699pa9-0'))
            )
            html_content = driver.page_source
        except Exception as e:
            logger.error(f"Failed to retrieve university home page: {url}. Error: {e}")
            return None
    else:
        try:
            driver.get(url)
            html_content = scroll_to_load_content(driver)
        except Exception as e:
            logger.error(f"Failed to retrieve the list of universities: {e}")
            return None

    cache[url] = html_content
    last_updated[url] = last_modified
    return html_content


def save_cache():
    with open(CACHE_FILE, 'wb') as f:
        pickle.dump(cache, f)
    return

def save_last_updated():
    with open(LAST_UPDATED_FILE, 'w') as f:
        f.write(json.dumps(last_updated))
    return


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        logger.error(f"An error occurred {e}. Exiting...")
    finally:
        driver.quit()
        helper.sys.exit(0)






from . import helper


def has_student_div(html_content):
    tag_name = 'h2'
    id_attr = 'studentbody'
    element = helper.get_html_element_by_id(html_content, tag_name, id_attr)
    if not element:
        return False
    return True

def get_student_data(html_content, about_us_paragraph):
    res = helper.OrderedDict({
        "enrollment": {
            "total": "N/A",
            "undergraduate": "N/A",
            "graduate": "N/A",
            "full_time_percentage": "N/A",
        },
        "demographics": get_demographics_data(html_content),
        "housing": get_housing_data(html_content),
        "campus_life": get_campus_life(html_content, about_us_paragraph)
    })

    if not has_student_div(html_content):
        return res

    divs = get_student_details_div(html_content)
    try:
        assert divs and len(divs) in [1, 2]
    except:
        return res

    try:
        under_grad_enrollment = divs[0].find('p', class_='StudentsSection__DataHeader-sc-1tb3548-0 dfCQxG')
        res["enrollment"]["undergraduate"] = under_grad_enrollment.text
    except:
        pass

    for inner_div in divs[1].find_all('div', class_='ContentSection__StyledDataRow-sc-699pa9-4'):
        paragraphs = inner_div.find_all('p', class_='Paragraph-sc-1iyax29-0')
        name = paragraphs[0].text
        value = paragraphs[1].text

        if 'Graduate Enrollment' in name:
            res["enrollment"]["graduate"] = value
        elif 'Total Enrollment' in name:
            res["enrollment"]["total"] = value
        elif 'Full-Time Degree' in name:
            try:
                value = (int(value.replace(',', '')) * 100)/int((res["enrollment"]["undergraduate"].replace(',', '')))
            except:
                pass
            else:
                res["enrollment"]["full_time_percentage"] = round(float(value), 2)

    return res


def get_student_details_div(html_content):
    h2_element = helper.get_html_element_by_id(html_content, 'h2', 'studentbody')
    if h2_element:
        next_div = h2_element.find_next_sibling('div', class_='Grid-lx2f3i-0 ContentSection__StyledGrid-sc-699pa9-1 fQYHzS cZfsa-D')
        if next_div:
            return next_div.find_all('div', class_='ContentSection__StyledCell-sc-699pa9-3')

    return None


def get_genders(html_content):
    divs = helper.get_html_elements(html_content, 'div', 'BarChartStacked__Legend-wgxhoq-4 iLdLaQ')
    if divs and len(divs) > 0:
        gender_divs = divs[0].find_all('div')
    else:
        return False
    male = 'N/A'
    female = 'N/A'

    for div in gender_divs:

        if 'male' in div.text.lower() and 'female' not in div.text.lower():
            try:
                male = div.find('b').text.rstrip('%')
            except:
                pass

        if 'female' in div.text.lower():
            try:
                female = div.find('b').text.rstrip('%')
            except:
                pass

    return {
        "male": male,
        "female": female
    }


def get_ethnicity(html_content):
    e = helper.OrderedDict({
        "white": "N/A",
        "asian": "N/A",
        "hispanic": "N/A",
        "black": "N/A",
        "international": "N/A",
        "unknown": "N/A",
        "two_or_more": "N/A",
        "native_american": "N/A",
        "pacific_islander": "N/A"
    })

    # Minority Enrollment 50
    # White 36
    # Asian 24
    # International 12
    # Hispanic 10
    # Black 9
    # Two or more races 7
    # Unknown 2

    divs = helper.get_html_elements(html_content, 'div', 'Key__ItemBox-sc-12afmmk-0')
    for div in divs:
        try:
            ethnicity = div.find('p', class_="Key__Title-sc-12afmmk-1").text.lower()
            inner_div = div.find('div', class_="Key__DataBox-sc-12afmmk-5")
            percent = inner_div.find('p', class_="Key__Data-sc-12afmmk-3").text.rstrip('%')
        except:
            continue
        if 'white' in ethnicity:
            e["white"] = percent
        elif 'asian' in ethnicity:
            e["asian"] = percent
        elif 'hispanic' in ethnicity:
            e["hispanic"] = percent
        elif 'black' in ethnicity:
            e["black"] = percent
        elif 'international' in ethnicity:
            e["international"] = percent
        elif 'unknown' in ethnicity:
            e["unknown"] = percent
        elif 'two or more' in ethnicity:
            e["two_or_more"] = percent

    return e

def get_demographics_data(html_content):
    genders = get_genders(html_content)

    demographics = {
        "gender": {
            "male": genders["male"] if genders else "N/A",
            "female": genders["female"] if genders else "N/A"
        },
        "ethnicity": get_ethnicity(html_content)
    }
    return demographics


def get_housing_data(html_content):
    h = {
        "students_in_college_housing": "N/A",
        "housing_types": []
    }

    housing_types = helper.get_html_elements(html_content, 'li', 'CampusLifeSection__StyledListItem-x6yma0-2')
    if not housing_types:
        return h

    for housing_type in housing_types:
        try:
            ht = housing_type.find('span').text
        except:
            continue
        h["housing_types"].append(ht)

    return h

def get_campus_life(html_content, about_us_paragraph):
    greek_life_elements = helper.get_html_elements(html_content, 'p', 'CampusLifeSection__DataHeader-x6yma0-8 lczeeQ')

    try:
        sports_team_count = greek_life_elements[0].text.strip('+')
        sports_team_count = int(sports_team_count)
    except:
        sports_team_count = 0

    try:
        fraternities = greek_life_elements[1].text
    except:
        fraternities = 'N/A'

    try:
        sororities = greek_life_elements[2].text
    except:
        sororities = 'N/A'

    campus_life = {
        "greek_life": {
            "has_greek_life": True if sports_team_count == 0 else False,
            "fraternities": fraternities,
            "sororities": sororities
        },
        "varsity_athletics": {
            "division": helper.get_answer(
                "What is the division of varsity athletics?",
                about_us_paragraph
            ),
            "men_teams": [helper.get_answer(
                "Varsity sports for men?",
                about_us_paragraph
            )],
            "women_teams": [helper.get_answer(
                "Varsity sports for women?",
                about_us_paragraph
            )]
        }
    }

    return campus_life


from . import helper


def get_admission_data(html_content):
    data = {
        "undergraduate": {
            "sat_score": 'N/A',
            "toefl_score": 'N/A',
        },
        "graduate": {
            "gmat_score": 'N/A',
            "ielts_score": 'N/A',
            "toefl_score": 'N/A',
        }
    }
    tag_name = 'a'
    id_attr = 'admission_Tab'
    elements = helper.get_html_element_by_id(html_content, tag_name, id_attr)
    if not elements:
        data

    tag_name = 'div'
    class_name = 'univ-subsection-full-width'
    elements = helper.get_html_elements(html_content, tag_name, class_name)

    undergrad_index = -1
    grad_index = -1
    for index in range(len(elements)):
        if 'Bachelor' in elements[index].text:
            undergrad_index = index
        if 'Master' in elements[index].text:
            grad_index = index

    if undergrad_index != -1:
        undergrad_div = elements[undergrad_index]
        outer_div = undergrad_div.find_all('div')
        inner_nested_div = outer_div[0].find_all('div')
        for index, item in enumerate(inner_nested_div):
            if 'SAT' in item.text:
                try:
                    sat_score_div = inner_nested_div[index+1]
                    data['undergraduate']['sat_score'] = sat_score_div.text
                except:
                    pass

            if 'TOEFL' in item.text:
                try:
                    toefl_score_div = inner_nested_div[index+1]
                    data['undergraduate']['toefl_score'] = toefl_score_div.text
                except:
                    pass

    if grad_index != -1:
        grad_div = elements[grad_index]
        outer_div = grad_div.find_all('div')
        inner_nested_div = outer_div[0].find_all('div')
        for index, item in enumerate(inner_nested_div):
            if 'GMAT' in item.text:
                try:
                    gmat_score_div = inner_nested_div[index+1]
                    data['graduate']['gmat_score'] = gmat_score_div.text
                except:
                    pass

            if 'IELTS' in item.text:
                try:
                    ielts_score_div = inner_nested_div[index+1]
                    data['graduate']['ielts_score'] = ielts_score_div.text
                except:
                    pass

            if 'TOEFL' in item.text:
                try:
                    toefl_score_div = inner_nested_div[index+1]
                    data['graduate']['toefl_score'] = toefl_score_div.text
                except:
                    pass

    return data





from . import helper


def has_ratings(html_content):
    tag_name = 'h2'
    id_attr = 'rankings'
    elements = helper.get_html_element_by_id(html_content, tag_name, id_attr)
    if not elements:
        return False
    return True

def get_over_all_rankings(html_content):
    tag_name = 'ul'
    class_name = 'RankList__List-sc-2xewen-0 ciVlaM util__RankList-sc-1kd04gx-3 dteFWJ'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    national_universities_rank = "N/A"
    best_value_schools_rank = "N/A"
    business_programs_rank = "N/A"
    other_ratings_url = "N/A"

    for li in elements[0]:
        a_tag = li.find('a')
        div = a_tag.find('div')
        rank_div = div.find('strong')
        if not rank_div:
            continue
        rank = rank_div.text

        types = a_tag.find_all('strong')
        other_types = []

        for type in types:
            if 'National Universities' in type.text:
                national_universities_rank = rank

            elif "Best Value Schools" in type.text:
                best_value_schools_rank = rank

            elif "Business Programs" in type.text:
                business_programs_rank = rank
            elif type.text != rank:
                other_types.append(f"{rank} in {type.text}")

        other_ratings_url = get_other_ratings_url(html_content)

    return (
        national_universities_rank,
        best_value_schools_rank,
        business_programs_rank,
        other_types,
    )


def get_other_ratings_url(html_content):
    tag_name = 'a'
    class_name = 'Anchor-byh49a-0 cXNfar'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    for element in elements:
        if element.get('data-tracking-placement') == 'Rankings':
            return element['href']
    return None


def get_ranking_data(html_content):
    data = {
        "national_universities_rank": "N/A",
        "best_value_schools_rank": "N/A",
        "business_programs_rank": "N/A",
        "other_rankings": ['N/A'],
    }
    if not has_ratings(html_content):
        helper.logger.info("No ratings found")
        return data

    national_universities_rank, best_value_schools_rank, business_programs_rank, \
        other_rankings = get_over_all_rankings(html_content)

    data['national_universities_rank'] = national_universities_rank
    data['best_value_schools_rank'] = best_value_schools_rank
    data['business_programs_rank'] = business_programs_rank
    data['other_rankings'] = other_rankings

    return data



# def get_qs_world_rankings(html_content):
#     try:
#         criteria_wrap = helper.get_html_elements(html_content, 'div', 'criteria-wrap')
#         circles = criteria_wrap[0].find_all('div', class_='circle')
#     except Exception as e:
#         helper.logger.error(f"Error in getting QS World Rankings {e}")
#         circles = []

#     return circles



from . import helper


def has_ratings(html_content):
    tag_name = 'div'
    id_attr = 'p2-rankings'
    elements = helper.get_html_element_by_id(html_content, tag_name, id_attr)
    if not elements:
        return False
    return True

def get_over_all_rankings(html_content):
    tag_name = 'a'
    class_name = 'use-ajax'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    rating = 'N/A'
    qs_world_rankings = 'N/A'
    qs_subject_ranking = 'N/A'
    qs_sustainability_ranking = 'N/A'
    for element in elements:
        try:
            rating = element.find('div', class_='latest_rank').text
        except:
            continue

        if element.get('id') == 'wur-tab':
            qs_world_rankings = rating
        elif element.get('id') == 'subj-tab':
            qs_subject_ranking = rating
        elif element.get('id') == 'item-4085':
            qs_sustainability_ranking = rating

    return qs_world_rankings, qs_subject_ranking, qs_sustainability_ranking


def get_ranking_data(html_content):
    data = {
        "qs_world_rankings": {
            "2025": {
                "overall_rank": "N/A",
                "scores": {
                    "overall": "N/A",
                    "academic_reputation": "N/A",
                    "citations_per_faculty": "N/A",
                    "employment_outcomes": "N/A",
                    "employer_reputation": "N/A",
                    "faculty_student_ratio": "N/A",
                    "international_faculty_ratio": "N/A",
                    "international_research_network": "N/A",
                    "international_student_ratio": "N/A",
                    "sustainability": "N/A"
                }
            }
        },
        "other_rankings": {
            "qs_subject_ranking": 'N/A',
            "qs_sustainability_ranking": 'N/A',
        }
    }
    if not has_ratings(html_content):
        helper.logger.info("No ratings found")
        return data

    qs_world_rankings, qs_subject_ranking, qs_sustainability_ranking = get_over_all_rankings(html_content)
    data['other_rankings']['qs_subject_ranking'] = qs_subject_ranking
    data['other_rankings']['qs_sustainability_ranking'] = qs_sustainability_ranking
    data['qs_world_rankings']['2025']['overall_rank'] = qs_world_rankings

    circles = get_qs_world_rankings(html_content)
    for circle in circles:
        try:
            score = circle.find('div', class_='score').text
            value = circle.find('div', class_='itm-name').text
            value = value.lower().replace(' ', '_')
            data['qs_world_rankings']['2025']['scores'][value] = score
        except Exception as e:
            helper.logger.error(f"Error in getting QS World Rankings {e}")
    return data



def get_qs_world_rankings(html_content):
    try:
        criteria_wrap = helper.get_html_elements(html_content, 'div', 'criteria-wrap')
        circles = criteria_wrap[0].find_all('div', class_='circle')
    except Exception as e:
        helper.logger.error(f"Error in getting QS World Rankings {e}")
        circles = []

    return circles



from . import helper


def get_admission_data(html_content):
    data = {
        "acceptance_rate": "N/A",
        "application_deadline": "N/A",
        "test_scores": {
            "sat": {
                "25th_percentile": "N/A",
                "75th_percentile": "N/A",
            },
            "act": {
                "25th_percentile": "N/A",
                "75th_percentile": "N/A",
            }
        },
        "high_school_gpa": {
            "reported": "N/A",
            "value": "N/A"
        }
    }

    adm_div = get_admission_details_div(html_content)
    if not adm_div:
        helper.logger.info("No admission details found")
        return data

    text_type = adm_div.find_all('div', class_='ApplyingSection__DataLabel-sc-2strss-8')
    try:
        assert text_type and text_type[0].text == 'Application Deadline'
        data["application_deadline"] = adm_div.find_all('p', class_='ApplyingSection__DataHeader-sc-2strss-7')[0].text
    except:
        pass

    try:
        assert text_type and text_type[1].text == 'Acceptance Rate'
        data["acceptance_rate"] = adm_div.find_all('p', class_='ApplyingSection__DataHeader-sc-2strss-7')[1].text
    except:
        pass

    scores = adm_div.find_all('div', class_='ApplyingSection__StyledDataRow-sc-2strss-5')
    for score in scores:
        paragraphs = score.find_all('p', class_='Paragraph-sc-1iyax29-0')
        if not paragraphs:
            continue

        if 'SAT' in paragraphs[0].text:
            try:
                sat_range = list(map(int, paragraphs[1].text.split('-')))
            except:
                continue
            else:
                data['test_scores']['sat']['25th_percentile'] = sat_range[0]
                data['test_scores']['sat']['75th_percentile'] = sat_range[1]

        elif 'ACT' in paragraphs[0].text:
            try:
                act_range = list(map(int, paragraphs[1].text.split('-')))
            except:
                continue
            else:
                data['test_scores']['act']['25th_percentile'] = act_range[0]
                data['test_scores']['act']['75th_percentile'] = act_range[1]

        elif 'High School GPA' in paragraphs[0].text:
            data['high_school_gpa']['reported'] = True
            data['high_school_gpa']['value'] = paragraphs[1].text


    return data


def get_admission_details_div(html_content):
    h2_element = helper.get_html_element_by_id(html_content, 'h2', 'admissions')
    if h2_element:
        next_div = h2_element.find_next_sibling('div', class_='Grid-lx2f3i-0 fQYHzS')
        if next_div:
            return next_div

    return None


from . import helper


def get_metadata(html_content, last_updated):
    try:
        data_source = helper.get_html_elements(html_content, 'span', 'Span-sc-19wk4id-0 Info__Copyright-sc-1vdhah7-9 lmLXYM jAudCl')[0]
        data_source = data_source.text
        data_source = data_source.split('©')[-1]
    except:
        data_source = 'N/A'

    return {
        "last_updated": last_updated,
        "data_source": data_source,
        "source_url": '',
        "scraping_timestamp": helper.get_current_time_stamp(),
    }




from . import helper


def get_student_data(html_content):
    students, international_students = get_student_count(html_content)
    percents = get_student_percents(html_content)
    res = helper.OrderedDict({
        "total_students": students,
        "undergraduate": {
            "percentage": percents['ug_percent'],
            "international_students": {
                "percentage": percents['ug_intr_percent'],
            }
        },
        "graduate": {
            "percentage": percents['pg_percent'],
            "international_students": {
                "percentage": percents['pg_intr_percent'],
            }
        },
        "total_international_students": international_students,
    })
    return res


def get_student_count(html_content):
    class_name = 'studstaff-subsection-count'
    tag_name = 'div'
    elements = helper.get_html_elements(html_content, tag_name, class_name)
    try:
        students = elements[0].text
    except:
        students = 'N/A'

    try:
        international_students = elements[1].text
    except:
        international_students = 'N/A'

    return students, international_students


def get_student_percents(html_content):
    parent_tag = 'div'
    parent_class = 'color-code-parent'
    nested_tag = 'div'

    elements = helper.get_html_elements(html_content, parent_tag, parent_class)

    try:
        ug_percent = helper.get_innermost_nested_element(elements[0], nested_tag)
    except:
        ug_percent = 'N/A'

    try:
        pg_percent = helper.get_innermost_nested_element(elements[1], nested_tag)
    except:
        pg_percent = 'N/A'

    try:
        ug_intr_percent = helper.get_innermost_nested_element(elements[2], nested_tag)
    except:
        ug_intr_percent = 'N/A'

    try:
        pg_intr_percent = helper.get_innermost_nested_element(elements[3], nested_tag)
    except:
        pg_intr_percent = 'N/A'

    res = {
        "ug_percent": ug_percent,
        "pg_percent": pg_percent,
        "ug_intr_percent": ug_intr_percent,
        "pg_intr_percent": pg_intr_percent
    }

    return res





import os
from setuptools import setup, find_packages

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

# Read requirements.txt
with open(os.path.join(BASE_DIR, "requirements.txt")) as f:
    install_requires = f.read().splitlines()

setup(
    name="ss-scrapping",
    version="0.1",
    packages=find_packages(),
    install_requires=install_requires,  # Use the dependencies from requirements.txt
    entry_points={
        "console_scripts": [
            "topuniversities = ss_scrapping.topuniversities.topuniversities:main",
            "usnews = ss_scrapping.usnews.usnews:main",
        ],
    },
    author="Parijat Srivastava",
    description="A Python package for top 500 universities",
)



from bs4 import BeautifulSoup
from collections import OrderedDict
import logging
import sys
import time
import datetime
import re

try:
    assert int(sys.argv[2]) == 1
except:
    pass
else:
    from transformers import pipeline

import os


NLP_MODEL = os.getenv("NLP_MODEL")
# NLP_MODEL = "distilbert/distilbert-base-cased-distilled-squad"
# NLP_MODEL = "bert-large-uncased-whole-word-masking-finetuned-squad"

logger = logging.getLogger(__name__)


def get_current_time_stamp():
    # format 2024-01-15T00:00:00Z
    return datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')

def get_html_elements(html_content, tag_name, class_name):
    elements = None
    try:
        soup = BeautifulSoup(html_content, 'html.parser')
        elements = soup.find_all(tag_name, class_=class_name)
    except Exception as e:
        logger.error(f"Error while parsing the html content: {e}")
    return elements

def get_innermost_nested_element(parent, nested_tag):
    nested_elements = parent.find_all(nested_tag)
    return (nested_elements[-1].text).rstrip('%')

def get_html_element_by_id(html_content, tag_name, id_name):
    soup = BeautifulSoup(html_content, 'html.parser')
    element = soup.find(tag_name, id=id_name)
    return element

def get_next_page(unordered_list, list_element):
    list_elements = unordered_list.find_all(list_element)
    list_element_anchor_tag = list_elements[-1].find('a')
    if not list_element_anchor_tag:
        return None
    next_page = list_element_anchor_tag['href'].lstrip('#')
    return next_page

def get_next_page_number_from_html_content(html_content):
    unordered_list = get_html_element_by_id(html_content, 'ul', 'alt-style-pagination')
    next_page = get_next_page(unordered_list, 'li')
    # next_page = page-2
    try:
        assert next_page
        next_page_number = int(next_page.split('-')[-1])
    except:
        return None
    return next_page_number

def get_count_of_universities_on_page(html_content):
    # gives the number written on homepage
    try:
        div = get_html_elements(html_content, 'div', 'filter-bar__CountContainer-sc-1glfoa-5')
        span = div[0].find('span')
        count = int(span.text)
    except:
        return 0
    return count

def get_university_list(html_content):
    tag_name = 'a'
    class_name = 'Card__StyledAnchor-sc-1ra20i5-10'
    elements = get_html_elements(html_content, tag_name, class_name=class_name)
    return elements

def get_universities_count(html_content):
    # gives the count/len of university elements found on the page
    try:
        tag_name = 'a'
        class_name = 'Card__StyledAnchor-sc-1ra20i5-10'
        elements = get_html_elements(html_content, tag_name, class_name=class_name)
        count = len(elements)
    except:
        return 0
    return count


def get_about_us_paragraph(html_content):
    class_name = 'MultilineEllipsis__FullHeightReferenceWrapper-sc-1hoyc1r-4'
    tag_name = 'div'
    elements = get_html_elements(html_content, tag_name, class_name)
    try:
        paragraph_ele = elements[0].find('p')
        nested_elements = paragraph_ele.find_all('p')
        paragraph = '\n'.join([element.text for element in nested_elements])
    except:
        paragraph = None

    return paragraph

def get_answer(question, paragraph):
    # print(question)
    # if question in [
    #     'How many active companies have alumni launched?',
    #     'How many jobs were created by alumni?',
    #     "How much annual revenue is generated by alumni's companies?",
    #     "What scientific breakthrough did research from the institution contribute to related to the expanding universe?"
    # ]:
    #     return 'N/A'
    try:
        assert int(sys.argv[2]) == 1
        assert paragraph and isinstance(paragraph, str) and len(paragraph) > 0
    except:
        return 'N/A'

    qa_pipeline = pipeline(
        "question-answering",
        model=NLP_MODEL
    )
    return qa_pipeline(
        question=question,
        context=paragraph
    ).get('answer', None)


def get_answer_in_list(question, paragraph):
    try:
        assert int(sys.argv[2]) == 1
    except:
        return ['N/A']

    try:
        answers = get_answer(question, paragraph)
        assert answers
        answers = [answers]
    except:
        answers = []

    return answers


# def check_page_end(html_content):
#     tag_name = 'div'
#     class_name = 'no-ranking-results-found'

#     elements = helper.get_html_elements(html_content, tag_name, class_name)
#     return bool(elements)


def get_university_name_and_url(element):
    url = element['href']
    h3 = element.find('h3')
    span = h3.find('span')
    name = f"{h3.text} {span.text}"
    return name, url
